from ..constants import __version__

ALL_LOGS_WE_PARSE = [
    "game.txt",
    "attack.txt",
    "pda.txt",
    "silicon.txt",
    "mecha.txt",
    "virus.txt",
    "telecomms.txt",
    "uplink.txt",
    "shuttle.txt"
]
HELP_LINK = "https://github.com/RigglePrime/SS13-tools/blob/master/ss13_tools/log_buddy/README.md"
SHAMELESS = f"## Created using SS13-Tools LogBuddy {__version__} https://github.com/RigglePrime/SS13-Tools\n"
ERRORED_FILE = "errored.log"
MAX_SUPPORTED_LOG_VERSION = "1.0.0"

LOG_COLOUR_SCARLET = 124
LOG_COLOUR_RED = 167
LOG_COLOUR_EMERALD = 35
LOG_COLOUR_PERIWINKLE = 69
LOG_COLOUR_PINK = 169
LOG_COLOUR_GRAY = 245
LOG_COLOUR_PASTEL_CYAN = 159
LOG_COLOUR_SUNSET = 209
LOG_COLOUR_PASTEL_ORANGE = 223
LOG_COLOUR_AMETHYST = 141
LOG_COLOUR_OCEAN = 27
